#!/usr/bin/env python3
#
#  project.py
"""
Comparison between projects.
"""
#
#  Copyright © 2024 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# 3rd party
import numpy
import pandas  # type: ignore[import-untyped]
from pyms.DPA.Alignment import Alignment

# this package
from libgunshotmatch.project import Project
from libgunshotmatch.utils import create_alignment

# this package
from ._utils import _get_padded_peak_list, _PaddedPeakList

__all__ = ("filter_alignment_to_consolidate", "get_padded_peak_list")


def filter_alignment_to_consolidate(project: Project) -> Alignment:
	"""
	Filter peaks from a project's ``alignment`` to only those which survived the ``consolidate`` process.

	:param project:
	"""

	assert project.consolidated_peaks is not None

	# Sort expr_code and peakpos into order from datafile_data
	desired_order = list(project.datafile_data)[::-1]
	sort_map = [desired_order.index(code) for code in project.alignment.expr_code]
	expr_code = [project.alignment.expr_code[idx] for idx in sort_map]
	peakpos = [project.alignment.peakpos[idx] for idx in sort_map]

	consolidated_peak_retention_times = [cp.rt_list for cp in project.consolidated_peaks]

	aligned_peaks_surviving_consolidate = []
	for aligned_peaks in zip(*peakpos):
		aprt = [p.rt for p in reversed(aligned_peaks)]
		if aprt in consolidated_peak_retention_times:
			aligned_peaks_surviving_consolidate.append(aligned_peaks)

	alignment_surviving_consolidate = list(zip(*aligned_peaks_surviving_consolidate))

	# Sanity check
	for c_expr_peaks, expr_peaks in zip(alignment_surviving_consolidate, peakpos):
		for peak in c_expr_peaks:
			assert peak in expr_peaks

	# Create new Alignment object
	return create_alignment(alignment_surviving_consolidate, expr_code)


def get_padded_peak_list(project: Project, alignment_rts: pandas.DataFrame) -> _PaddedPeakList:
	"""
	Returns a list of consolidated peaks for the project, based on the between-project alignment.

	:param project:
	:param alignment_rts: Pandas DataFrame giving retention times for the peak alignment.
		The output of :meth:`~pyms.DPA.Alignment.Alignment.get_peak_alignment`.

	"""

	rts = [numpy.mean(row[1:]) for row in alignment_rts[list(project.datafile_data)].itertuples()]

	return _get_padded_peak_list(project, rts)



#!/usr/bin/env python3
#
#  utils.py
"""
Utility functions.
"""
#
#  Copyright © 2020-2023 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Iterable, List, Optional, Sequence, Tuple, Type, TypeVar, Union

# 3rd party
import numpy
from attr import AttrsInstance
from chemistry_tools.spectrum_similarity import SpectrumSimilarity
from mathematical.utils import rounders
from pyms.DPA.Alignment import Alignment
from pyms.Peak.Class import Peak
from pyms.Spectrum import MassSpectrum
from scipy.stats import truncnorm  # type: ignore[import-untyped]

if TYPE_CHECKING:
	# this package
	from libgunshotmatch.project import Project

__all__ = ("round_rt", "get_truncated_normal", "ms_comparison", "get_rt_range", "create_alignment")


def round_rt(rt: Union[str, float, Decimal]) -> Decimal:
	"""
	Truncate precision of retention time to 10 decimal places.

	:param rt:
	"""

	# Limit to 10 decimal places as that's what Pandas writes JSON data as;
	# no need for greater precision.

	return rounders(rt, "0.0000000000")


def get_truncated_normal(
		mean: float,
		sd: float,
		low: float = 0,
		upp: float = 10,
		count: int = 10,
		random_state: Optional[int] = None,
		) -> Sequence[float]:
	"""
	Returns ``count`` values from a truncated normal distrubition.

	:param mean: The midpoint of the normal distribution.
	:param sd: The spread of the normal distribution (the standard deviation).
	:param low: The lower bound.
	:param upp: The upper bound.
	:param count:
	:param random_state: Optional seed for the random number generator.
	"""

	# From https://stackoverflow.com/a/74448424
	# By toco_tico https://stackoverflow.com/users/1060349/toto-tico
	# CC BY-SA 4.0

	dist = truncnorm((low - mean) / sd, (upp - mean) / sd, loc=mean, scale=sd)
	return dist.rvs(count, random_state=random_state)


def ms_comparison(top_ms: MassSpectrum, bottom_ms: MassSpectrum) -> Optional[float]:
	"""
	Performs a Mass Spectrum similarity calculation two mass spectra.

	:param top_ms:
	:param bottom_ms:

	If either of ``top_ms`` or ``bottom_ms`` is :py:obj:`None` then :py:obj:`None` is returned,
	otherwise a comparison score is returned.
	"""

	if top_ms is None or bottom_ms is None:
		return None

	top_spec = numpy.column_stack((top_ms.mass_list, top_ms.mass_spec))
	bottom_spec = numpy.column_stack((bottom_ms.mass_list, bottom_ms.mass_spec))

	sim = SpectrumSimilarity(
			top_spec,
			bottom_spec,
			b=1,
			xlim=(45, 500),  # TODO: configurable or taken from spectra
			)

	match, rmatch = sim.score()
	return match * 1000


_AI = TypeVar("_AI", bound=AttrsInstance)


def _fix_init_annotations(method: Type[_AI]) -> Type[_AI]:
	init_annotations = method.__init__.__annotations__
	cls_annotations = method.__annotations__

	for k, v in cls_annotations.items():
		if k in init_annotations:
			if init_annotations[k] is Any:
				init_annotations[k] = v
		else:
			init_annotations[k] = v

	return method


def _to_list(l: Iterable[str]) -> List[str]:
	"""
	Attrs type hint helper for converting to a list.

	Otherwise the errors are:

	libgunshotmatch/consolidate/__init__.py:701: error: Argument "name_filter" to "ConsolidatedPeakFilter" has incompatible type "List[str]"; expected "Iterable[_T]"  [arg-type]
	libgunshotmatch/project.py:202: error: List item 0 has incompatible type "str"; expected "_T"  [list-item]
	"""

	return list(l)


def get_rt_range(project: "Project") -> Tuple[float, float]:
	"""
	Returns the minimum and maximum retention times (in minutes) across the repeats.

	:param project:

	:rtype:

	.. versionadded:: 0.7.0
	"""

	# Get RT extremes from intensity matrix
	min_rts, max_rts = [], []
	for repeat in project.datafile_data.values():
		im = repeat.datafile.intensity_matrix
		assert im is not None
		times = im.time_list
		min_rts.append(times[0])
		max_rts.append(times[-1])

	min_rt = min(min_rts) / 60
	max_rt = max(max_rts) / 60

	return min_rt, max_rt


def create_alignment(
		peakpos: Sequence[Sequence[Optional[Peak]]],
		expr_code: List[str],
		similarity: float = 0,
		) -> Alignment:
	"""
	Create a new :class:`pyms.DPA.Alignment.Alignment` object.

	:param peakpos: Nested list of aligned peaks. Top level list contains lists of peaks for each experiment in ``expr_code``.
	:param expr_code: Experiment names. Order must match ``peakpos``.
	:param similarity:

	:rtype:

	.. versionadded:: 0.8.0
	"""

	alignment = Alignment(None)
	alignment.peakpos = [list(p) for p in peakpos]  # type: ignore[arg-type]
	alignment.peakalgt = numpy.transpose(alignment.peakpos).tolist()  # type: ignore[arg-type]
	alignment.expr_code = expr_code
	alignment.similarity = similarity

	return alignment



#!/usr/bin/env python3
#
#  _utils.py
"""
Utility functions.
"""
#
#  Copyright © 2024 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from typing import List, MutableSequence, Optional

# 3rd party
import numpy

# this package
from libgunshotmatch import project
from libgunshotmatch.consolidate import ConsolidatedPeak

_PaddedPeakList = MutableSequence[Optional[ConsolidatedPeak]]


def _get_padded_peak_list(project: project.Project, rts: List[float]) -> _PaddedPeakList:

	assert project.consolidated_peaks is not None

	padded_cp_list: _PaddedPeakList = []

	for cp in project.consolidated_peaks:
		if rts:
			top_rt = rts.pop(0)
			while numpy.isnan(top_rt):
				padded_cp_list.append(None)
				if rts:
					top_rt = rts.pop(0)
				else:
					break
		padded_cp_list.append(cp)

	return padded_cp_list



#!/usr/bin/env python3
#
#  _fields.py
"""
Internal attrs field helpers.
"""
#
#  Copyright © 2023 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from typing import Any, Generic, Iterable, Optional, Set, Tuple, Type, TypeVar, Union, overload

# 3rd party
import attr
from pyms.Utils.Time import time_str_secs

__all__ = (
		"Boolean",
		"FieldType",
		"Integer",
		"Number",
		"String",
		"convert_crop_mass_range",
		"convert_rt_range",
		"default_base_peak_filter"
		)

_FT = TypeVar("_FT")


class FieldType(Generic[_FT]):
	"""
	Customisable method field type.
	"""

	#: The Python type of the object.
	field_type: Type[_FT]

	#: String name of the field.
	field_name: str

	def __new__(cls: Type["FieldType"]) -> Type["FieldType"]:  # type: ignore[misc]
		"""
		Make the ``FieldType`` class unitializable.

		Mypy doesn't like it, hence the ``type: ignore``.
		"""

		return cls

	@classmethod
	def validator(cls: Type["FieldType"], inst: object, attr: attr.Attribute, value: Any) -> None:
		"""
		Check if a value conforms to this field's expected datatype.

		This method is called by attrs.
		"""

		if not isinstance(value, cls.field_type):
			err_msg = f"'{attr.name}' must be {cls.field_name} (got {value!r} that is a {value.__class__!r})."
			raise TypeError(err_msg)

	@classmethod
	def field(cls: Type["FieldType"], default: _FT) -> _FT:
		"""
		Construct an attrs field.
		"""
		# Actually returns attr.Attribute, but mypy doesn't like it

		return attr.field(
				default=default,
				converter=cls.field_type,
				validator=cls.validator,
				)


class Boolean(FieldType):
	"""
	Boolean method field type.
	"""

	field_type = bool
	field_name = "boolean"


class String(FieldType):
	"""
	String method field type.
	"""

	field_type = str
	field_name = "a string"


class Integer(FieldType):
	"""
	Integer method field type.
	"""

	field_type = int
	field_name = "an integer"


class Number(FieldType):
	"""
	Numerical method field type.
	"""

	field_type = float
	field_name = "a number"

	@classmethod
	def validator(cls: Type["FieldType"], inst: object, attr: attr.Attribute, value: Any) -> None:
		"""
		Check if a value is a number.

		This method is called by attrs.
		"""

		if not isinstance(value, (int, float)):
			err_msg = f"'{attr.name}' must be {cls.field_name} (got {value!r} that is a {value.__class__!r})."
			raise TypeError(err_msg)


def convert_crop_mass_range(value: Optional[Iterable]) -> Optional[Tuple[int, int]]:
	"""
	Convert a value to a tuple of integers for the ``crop_mass_range`` option.
	"""

	if not value:
		return None

	value_as_tuple = tuple([int(x) for x in value])
	if len(value_as_tuple) != 2:
		err_msg = f"crop_mass_range must be a 2-element tuple. Got {len(value_as_tuple)} values: {value_as_tuple}"
		raise ValueError(err_msg)

	return value_as_tuple


@overload
def convert_sg_window(value: str) -> str: ...


@overload
def convert_sg_window(value: int) -> int: ...


def convert_sg_window(value: Union[str, int]) -> Union[str, int]:
	"""
	Convert a Savitzky-Golay window size parameter value.
	"""

	if isinstance(value, int):
		return value
	elif isinstance(value, str):
		# Check the string can be correctly parsed
		time_str_secs(value)
		return value
	else:
		raise TypeError("'savitzky_golay_window' must be either an integer or a string")


def convert_rt_range(value: Optional[Iterable]) -> Optional[Tuple[float, float]]:
	"""
	Convert a value to a tuple of floats for the ``rt_range`` option.
	"""

	if not value:
		return None

	value_as_tuple = tuple([float(x) for x in value])
	if len(value_as_tuple) != 2:
		raise ValueError(f"rt_range must be a 2-element tuple. Got {len(value_as_tuple)} values: {value_as_tuple}")

	return value_as_tuple


def default_base_peak_filter() -> Set[int]:
	"""
	Returns the default value for the ``base_peak_filter`` option.
	"""

	return {73, 147}



#!/usr/bin/env python3
#
#  peak.py
"""
Classes representing peaks, and functions for peak filtering.
"""
#
#  Copyright © 2020-2023 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from typing import Any, Collection, Dict, List, Mapping, Optional, Sequence, Type, Union

# 3rd party
import numpy
import pandas  # type: ignore[import-untyped]
import sdjson
from domdf_python_tools.paths import PathPlus
from domdf_python_tools.typing import PathLike
from pyms.BillerBiemann import num_ions_threshold
from pyms.DPA.Alignment import exprl2alignment
from pyms.DPA.PairwiseAlignment import Alignment, PairwiseAlignment, align_with_tree
from pyms.Experiment import Experiment
from pyms.IonChromatogram import IonChromatogram
from pyms.Noise.Analysis import window_analyzer
from pyms.Peak.Class import AbstractPeak, ICPeak, Peak
from pyms.Peak.List.Function import sele_peaks_by_rt
from pyms.Spectrum import MassSpectrum
from pyms_nist_search import SearchResult

__all__ = (
		"PeakList",
		"QualifiedPeak",
		"QualifiedPeakList",
		"align_peaks",
		"filter_aligned_peaks",
		"filter_peaks",
		"peak_from_dict",
		"write_alignment",
		"base_peak_mass",
		)


class QualifiedPeak(Peak):
	"""
	A Peak that has been identified using NIST MS Search and contains a list of possible identities.

	:param rt: Retention time.
	:param ms: The mass spectrum at the apex of the peak.
	:param minutes: Retention time units flag. If :py:obj:`True`, retention time
		is in minutes; if :py:obj:`False` retention time is in seconds.
	:param outlier: Whether the peak is an outlier.
	:param hits: List of possible identities for the peak.
	:param peak_number: Optional numerical identifier for the :class:`~pyms.Peak.Class.Peak`, such as in an :class:`~.pyms.DPA.Alignment.Alignment`.
	"""

	#: List of possible identities for the peak.
	hits: List[SearchResult]
	#: Optional numerical identifier for the peak, such as in an :class:`~.pyms.DPA.Alignment.Alignment`.
	peak_number: Optional[int]

	def __init__(
			self,
			rt: float = 0.0,
			ms: Optional[MassSpectrum] = None,
			minutes: bool = False,
			outlier: bool = False,
			hits: Optional[List[SearchResult]] = None,
			peak_number: Optional[int] = None
			):

		Peak.__init__(self, rt, ms, minutes, outlier)

		if hits is not None:
			if not isinstance(hits, list) or not isinstance(hits[0], SearchResult):
				raise TypeError("'hits' must be a list of SearchResult objects")

		if hits is None:
			self.hits = []
		else:
			self.hits = hits

		self.peak_number = peak_number

	def __new__(  # noqa: D102
		cls,
		rt: float = 0.0,
		ms: Optional[MassSpectrum] = None,
		minutes: bool = False,
		outlier: bool = False,
		hits: Optional[List[SearchResult]] = None,
		peak_number: Optional[int] = None
		):
		# Overrides __new__ method which warns if the Peak is for an IC (not applicable here)

		obj = object.__new__(cls)
		obj.__init__(rt, ms, minutes, outlier, hits, peak_number)  # type: ignore[misc]
		return obj

	def __eq__(self, other) -> bool:  # noqa: MAN001
		"""
		Return whether this QualifiedPeak object is equal to another object.

		:param other: The other object to test equality with.
		"""

		if isinstance(other, self.__class__):
			return (
					self.UID == other.UID and self.bounds == other.bounds and self.rt == other.rt
					and self.mass_spectrum == other.mass_spectrum and self.area == other.area
					and self.hits == other.hits and self.peak_number == other.peak_number
					)

		return NotImplemented

	@classmethod
	def from_peak(cls: Type["QualifiedPeak"], peak: Peak) -> "QualifiedPeak":
		"""
		Construct :class:`~.QualifiedPeak` from a :class:`~pyms.Peak.Class.Peak`.

		The resulting :class:`~.QualifiedPeak` will not have :attr:`~.hits` or :attr:`~.peak_number` set,
		but those attributes can be set after calling this method.

		:param peak:
		"""

		if not isinstance(peak, AbstractPeak):
			raise TypeError("'peak' must be a Peak object")

		if isinstance(peak, ICPeak):
			new_peak = cls(peak.rt, peak.ic_mass, False, peak.is_outlier)  # type: ignore[arg-type]  # TODO
		else:
			new_peak = cls(peak.rt, peak.mass_spectrum, False, peak.is_outlier)

		new_peak.area = peak.area
		new_peak.bounds = peak.bounds
		new_peak._UID = peak.UID

		return new_peak

	def __repr__(self) -> str:
		return f"<Qualified Peak: {self.rt}>"

	def __str__(self) -> str:
		return self.__repr__()

	def to_dict(self) -> Dict[str, Any]:
		"""
		Returns a dictionary representation of this peak.

		All keys are native, JSON-serializable, Python objects.
		"""

		try:
			ion_areas: Union[Dict, float] = self.ion_areas
		except ValueError:
			ion_areas = 0

		return {
				"UID": self.UID,
				"area": self.area,
				"bounds": self.bounds,
				"ion_areas": ion_areas,
				"is_outlier": self.is_outlier,
				"mass_spectrum": {
						"intensity_list": self.mass_spectrum.intensity_list,
						"mass_list": self.mass_spectrum.mass_list,
						},
				"rt": self.rt,
				"hits": [hit.to_dict() for hit in self.hits],
				"peak_number": self.peak_number,
				}

	@classmethod
	def from_dict(cls: Type["QualifiedPeak"], d: Mapping[str, Any]) -> "QualifiedPeak":
		"""
		Construct a :class:`~.QualifiedPeak` from a dictionary.

		:param d:
		"""

		hits = [SearchResult.from_dict(hit) for hit in d["hits"]]
		peak_obj = QualifiedPeak(
				d["rt"],
				MassSpectrum(**d["mass_spectrum"]),
				outlier=d["is_outlier"],
				hits=hits,
				peak_number=d["peak_number"]
				)
		peak_obj.bounds = d["bounds"]
		peak_obj._UID = d["UID"]
		peak_obj.area = d["area"]
		if d["ion_areas"]:
			peak_obj.ion_areas = d["ion_areas"]

		return peak_obj


# @prettify_docstrings
class PeakList(List[Peak]):
	"""
	Represents a list of peaks.

	.. autosummary-widths:: 35/100
	"""

	#: String identifier for the datafile the peaks were detected in.
	datafile_name: Optional[str] = None

	def __repr__(self) -> str:
		if self.datafile_name:
			return f"{self.__class__.__name__}(datafile={self.datafile_name}; <{len(self)} peaks>)"
		else:
			return f"{self.__class__.__name__}(<{len(self)} peaks>)"

	def __str__(self) -> str:
		return self.__repr__()

	def to_list(self) -> List[Dict[str, Any]]:
		"""
		Return a list of pure-Python dictionaries representing the peaks and their mass spectra.
		"""

		peaks_as_pure_list: List[Dict[str, Any]] = []
		for peak in self:
			if peak is None:
				peaks_as_pure_list.append(None)
				continue

			try:
				ion_areas = peak.ion_areas
			except ValueError:
				ion_areas = None

			peaks_as_pure_list.append({
					"UID": peak.UID,
					"area": peak.area,
					"bounds": peak.bounds,
					"ion_areas": ion_areas,
					"is_outlier": peak.is_outlier,
					"mass_spectrum": {
							"intensity_list": peak.mass_spectrum.intensity_list,
							"mass_list": peak.mass_spectrum.mass_list,
							},
					"rt": peak.rt,
					})

		return peaks_as_pure_list


# @prettify_docstrings
class QualifiedPeakList(List[QualifiedPeak]):
	"""
	Represents a list of qualified peaks.
	"""

	#: String identifier for the datafile the peaks were detected in.
	datafile_name: Optional[str] = None

	def __repr__(self) -> str:
		if self.datafile_name:
			return f"{self.__class__.__name__}(datafile={self.datafile_name}; <{len(self)} peaks>)"
		else:
			return f"{self.__class__.__name__}(<{len(self)} peaks>)"

	def __str__(self) -> str:
		return self.__repr__()

	def to_list(self) -> List[Dict[str, Any]]:
		"""
		Return a list of pure-Python dictionaries representing the peaks and their mass spectra.
		"""

		return [qp.to_dict() for qp in self]


def filter_peaks(
		peak_list: List[Peak],
		tic: IonChromatogram,
		noise_filter: bool = True,
		noise_threshold: int = 2,
		base_peak_filter: Collection[int] = (73, 147),
		rt_range: Optional[Sequence[float]] = None,
		) -> PeakList:
	"""
	Filter a list of peaks to remove noise and peaks due to e.g. column bleed.

	:param peak_list:
	:param tic: The TIC of the GC-MS data from which these peaks were identified.
	:param noise_filter: Whether to perform automatic noise filtering of the peak list.
	:param noise_threshold: The minimum number of ions that must have intensities above the noise floor, otherwise the peak is excluded.
	:param base_peak_filter: Peaks whose base peak is at one of the listed masses (*m/z*) are excluded.
	:param rt_range: Optional retention time range (in minutes) to filter the peak list to.
	"""

	if noise_filter:
		# Filtering peak lists with automatic noise filtering
		noise_level = window_analyzer(tic)
		# should we also do rel_threshold() here?
		# https://pymassspec.readthedocs.io/en/master/pyms/BillerBiemann.html#pyms.BillerBiemann.rel_threshold
		peak_list = num_ions_threshold(peak_list, noise_threshold, noise_level)

	if rt_range is not None:
		# Crop time range
		peak_list = PeakList(sele_peaks_by_rt(peak_list, [f"{rt_range[0]}m", f"{rt_range[1]}m"]))

	final_peak_list = PeakList()

	for peak in peak_list:
		# Get mass and intensity lists for the mass spectrum at the apex of the peak
		apex_mass_list = peak.mass_spectrum.mass_list
		apex_mass_spec = peak.mass_spectrum.mass_spec

		# Determine the intensity of the base peak in the mass spectrum
		base_peak_intensity = max(apex_mass_spec)

		# Determine the index of the base peak in the mass spectrum
		base_peak_index = [
				index for index, intensity in enumerate(apex_mass_spec) if intensity == base_peak_intensity
				][0]

		# Finally, determine the mass of the base peak
		base_peak_mass = apex_mass_list[base_peak_index]

		# skip the peak if the base peak is at e.g. m/z 73, i.e. septum bleed
		if base_peak_mass in base_peak_filter:  # method.base_peak_filter:
			continue

		final_peak_list.append(peak)

	return final_peak_list


@sdjson.register_encoder(numpy.int64)
def _encode_numpy_int64(obj: numpy.int64) -> int:
	# Necessary as to_dict() sometimes includes int64 which isn't natively serializable.
	return int(obj)


def align_peaks(
		peaks: List[PeakList],
		rt_modulation: float = 2.5,
		gap_penalty: float = 0.3,
		min_peaks: int = 1,
		) -> Alignment:
	"""
	Perform peak alignment.

	:param peaks: List of list of identified peaks. Each :class:`~.PeakList` must have its :attr:`~.PeakList.datafile_name` attribute set.
	:param rt_modulation: Retention time tolerance parameter for pairwise alignments.
	:param gap_penalty: Gap parameter for pairwise alignments.
	:param min_peaks: Minimum number of peaks required for the alignment position to survive filtering.
		If set to ``-1`` the number of repeats in the project are used.

	:rtype:

	.. latex:clearpage::
	"""

	print("\nAligning\n")

	expr_list = []
	for peak_list in peaks:
		if peak_list.datafile_name is None:
			raise ValueError("Cannot align peaks with PeakList.datafile_name unset")
		else:
			expr_list.append(Experiment(peak_list.datafile_name, peak_list))

	F1: List[Alignment] = exprl2alignment(expr_list)
	# F1: List[Alignment] = exprl2alignment([Experiment(d["datafile"].name, d["peak_list"]) for d in expr_list])

	T1 = PairwiseAlignment(
			F1,
			rt_modulation,
			gap_penalty,
			)

	if min_peaks == -1:
		min_peaks = len(expr_list)

	A1: Alignment = align_with_tree(T1, min_peaks=min_peaks)

	return A1


def write_alignment(
		alignment: Alignment,
		project_name: str,
		output_dir: PathLike,
		require_all_datafiles: bool = False,
		) -> None:
	"""
	Write the alignment data (retention times, peak areas, mass spectra) to disk.

	The output files are as follows:


	* :file:`{{project_name}}_alignment_rt.csv`, containing the aligned retention times.
	* :file:`{{project_name}}_alignment_area.csv`, containing the peak areas for the corresponding aligned retention times.
	* :file:`{{project_name}}_alignment_rt.json`, containing the aligned retention times.
	* :file:`{{project_name}}_alignment_area.json`, containing the peak areas for the corresponding aligned retention times.
	* :file:`{{project_name}}_alignment_ms.json`, containing the mass spectra for the corresponding aligned retention times.

	:param alignment:
	:param project_name: The name of the project. Prefixed to all filenames.
	:param output_dir: Directory to store the output files in.
	:param require_all_datafiles: Whether the peak must be present in all experiments to be included in the data frame.
	"""

	output_dir_p = PathPlus(output_dir)

	alignment.write_csv(
			output_dir_p / f"{project_name}_alignment_rt.csv",
			output_dir_p / f"{project_name}_alignment_area.csv",
			)

	rt_alignment = alignment.get_peak_alignment(require_all_expr=require_all_datafiles)
	rt_alignment_filename = output_dir_p / f"{project_name}_alignment_rt.json"
	rt_alignment_filename.write_clean(rt_alignment.to_json())

	area_alignment = alignment.get_area_alignment(require_all_expr=require_all_datafiles)
	area_alignment_filename = output_dir_p / f"{project_name}_alignment_area.json"
	area_alignment_filename.write_clean(area_alignment.to_json())

	ms_alignment = alignment.get_ms_alignment(require_all_expr=require_all_datafiles)
	# ms_alignment.to_json(output_dir_p / 'alignment_ms.json')
	alignment_ms_filename = (output_dir_p / f"{project_name}_alignment_ms.json")
	alignment_ms_filename.dump_json(
			ms_alignment.to_dict(),
			json_library=sdjson,  # type: ignore[arg-type]
			)


def filter_aligned_peaks(
		alignment: Alignment,
		top_n_peaks: int = 80,
		min_peak_area: float = 0,
		) -> pandas.DataFrame:
	"""
	Filter aligned peaks by minimum average peak area, and to the top ``n`` largest peaks.

	:param alignment:
	:param top_n_peaks: Filter to the largest ``n`` peaks. If ``0`` all peaks are included.
	:param min_peak_area: Exclude aligned peaks with an average peak area below this threshold.

	:returns: :class:`pandas.DataFrame` giving the retention times of the aligned peaks.
	"""

	# Get peak area and retention times from Alignment
	area_alignment: pandas.DataFrame = alignment.get_area_alignment(require_all_expr=False)
	rt_alignment: pandas.DataFrame = alignment.get_peak_alignment(require_all_expr=False)

	# Calculate average peak area for each of the aligned peaks
	area_alignment["mean"] = area_alignment[alignment.expr_code].mean(axis=1)
	area_alignment["stdev"] = area_alignment[alignment.expr_code].std(axis=1)

	# print(area_alignment[["mean", "stdev"]])

	# Sort mean_peak_areas from largest (top) to smallest
	area_alignment = area_alignment.sort_values(by="mean")

	########
	# Get indices of largest n peaks based on `ident_top_peaks`
	top_peaks_indices = []

	if top_n_peaks:
		print(f"Filtering to the largest {top_n_peaks} peaks with an average peak area above {min_peak_area}")

		# print("tail of area_alignment=", area_alignment.tail(top_n_peaks))

		# Limit to the largest `top_n_peaks` peaks
		for peak_no, areas in area_alignment.tail(top_n_peaks).iterrows():
			# Ignore peak if average peak area is less then min_peak_area
			if areas["mean"] >= min_peak_area:
				top_peaks_indices.append(peak_no)

	else:
		print(f"Filtering to peaks with an average peak area above {min_peak_area}")
		for peak_no, areas in area_alignment.iterrows():
			# Ignore peak if average peak area is less then min_peak_area
			if areas["mean"] >= min_peak_area:
				top_peaks_indices.append(peak_no)

	# Remove peaks from rt_alignment if they are not in top_peaks_indices,
	# i.e. they are one of the largest n largest peaks
	rt_alignment = rt_alignment.filter(top_peaks_indices, axis=0)

	return rt_alignment


def peak_from_dict(d: Dict[str, Any]) -> Peak:
	"""
	Construct a :class:`~pyms.Peak.Class.Peak` from a dictionary.

	:param d:

	:rtype:

	.. latex:clearpage::
	"""

	peak_obj = Peak(d["rt"], MassSpectrum(**d["mass_spectrum"]), outlier=d["is_outlier"])
	peak_obj.bounds = d["bounds"]
	peak_obj._UID = d["UID"]
	peak_obj.area = d["area"]
	if d["ion_areas"]:
		peak_obj.ion_areas = d["ion_areas"]

	return peak_obj


def _to_peak_list(a_list: List[Peak]) -> PeakList:
	"""
	Internal utility to coerce a list of peaks to an actual :class:`~.PeakList`.
	"""

	if isinstance(a_list, PeakList):
		return a_list
	else:
		return PeakList(a_list)


def base_peak_mass(peak: Peak) -> float:
	"""
	Returns the mass of the largest fragment in the peak's mass spectrum.

	:param peak:

	.. versionadded:: v0.11.0
	"""

	apex_mass_list = peak.mass_spectrum.mass_list
	apex_mass_spec = peak.mass_spectrum.mass_spec

	# Determine the intensity of the base peak in the mass spectrum
	base_peak_intensity = max(apex_mass_spec)

	# Determine the index of the base peak in the mass spectrum
	base_peak_index = apex_mass_spec.index(base_peak_intensity)

	# Finally, determine the mass of the base peak
	return apex_mass_list[base_peak_index]
